#!/usr/bin/python

import sys
import pygtk #; pygtk.require("2.0")
import gobject
import gtk
import gtk.glade
import pango
import os.path
import os
# import configobj
import sys
from config import *
from ButtonLayout import *

# platform-independent functions

# def getdatadir():
#     if hasattr(sys, "frozen"):                 # windows
#         return os.path.join(os.path.dirname(sys.executable), "data")
#     else:
#         return os.path.join(os.path.dirname(sys.argv[0]), "..", "data")

# def lastword(text, symbol=" "):
#     return text[text.rfind(symbol) + 1:]

def listitems(list):
    for i in range(len(list)):
        yield i, list[i]

def prex(expression, prefix=""):
    print prefix, expression
    return expression

# def cutcomment(text):
#     pos = text.find("#")
#     if pos < 0:
#         return text
#     else:
#         return text[:pos]

def reversed(list):
    return list[::-1]

def recursewidgets(widget, callback):
    callback(widget)
    if isinstance(widget, gtk.Container):
        for child in widget.get_children():
            recursewidgets(child, callback)

def marktoiter(mark):
    return mark.get_buffer().get_iter_at_mark(mark)

# def addtoiter(iter, count):
#     iter = iter.copy()
#     iter.forward_chars(count)
#     return iter

def gtkdowork():
    while gtk.events_pending():
        gtk.main_iteration()

# Todo:
# - make font selection style based: gtk.RcStyle, set_style

class MyApp:
    def __init__(self):
        # setup state variables
        self.internal = False
        self.wordlistprefix = None
        self.individualmode = False
        self.worddict = None
        self.iterhistory = []

        # load layout data
        gtk.rc_parse("tapirrc.gtk")           
        self.config = TapirConfig(self)
        self.tree = gtk.glade.XML("tapir.glade")

        # set widgets
        self.mainwindow = self.widget("MainWindow")
        self.buttontable = self.widget("ButtonTable")
        self.spacingscale = self.widget("SpacingScale")
        self.wordlist = self.widget("wordlist")
        self.textview = self.widget("textview")
        self.mainwindow.set_property("allow-shrink", True)

        self.layout = ButtonLayout("layout.xml", self.tree, self.mainwindow, self)

        # set up the textview and textbuffer
        self.textbuf = gtk.TextBuffer()
        self.textview.set_buffer(self.textbuf)
        self.badpretag = self.textbuf.create_tag(background="#ffaaaa")
        self.pretag = self.textbuf.create_tag(background="#aaffaa")
        self.posttag = self.textbuf.create_tag(background="#aaaaff")

        # set up the word list
        self.liststore = gtk.ListStore(gobject.TYPE_STRING)
        self.wordlist.set_model(self.liststore)

        cell = gtk.CellRendererText()
        tvcolumn = gtk.TreeViewColumn()
        tvcolumn.pack_start(cell)
        tvcolumn.add_attribute(cell, 'text', 0)

        self.wordlist.append_column(tvcolumn)

        # set up the marks
        self.cursor = self.textbuf.get_insert()
        self.start = self.textbuf.create_mark(None, marktoiter(self.cursor),
                                              True)
        self.end = self.textbuf.create_mark(None, marktoiter(self.cursor))

        # setup an invisible cursor
        pix = gtk.gdk.Pixmap(self.buttontable.window, 1, 1, 1)
        color = gtk.gdk.Color()
        self.nullcursor = gtk.gdk.Cursor(pix, pix, color, color, 0, 0)
        self.normalcursor = gtk.gdk.Cursor(gtk.gdk.TOP_LEFT_ARROW)
        
        # load data
        self.preparedictionary()
        self.suffixdict = self.loaddictionarygraphically("suffixes.dict")
        self.speech = os.popen("festival --pipe", "w")

#         symbols = "abc"
#         abcbox = gtk.VBox()
#         for word in self.worddict["2"][:7]:
#             if word not in symbols:
#                 label = gtk.Label(word)
#                 label.set_alignment(0, 0.5)
#                 abcbox.pack_start(label)
#         mybutton = gtk.HBox()
#         mybutton.pack_start(abcbox)
#         mybutton.pack_start(gtk.Label(symbols))

#         abc = self.widget("button13")
#         abc.remove(abc.get_child())
#         abc.add(mybutton)


        # connect events
        self.textbuf.connect("mark_set", self.onmarkset)
        self.textbuf.connect("changed", self.onchanged)
        self.tree.signal_autoconnect(self)
        self.buttontable.realize()
        self.mainwindow.realize()
#        print "booo"
        self.config.load()
        self.mainwindow.show_all()
        self.config.loadattr("hidemouse")
#        self.mainwindow.set_geometry_hints(None, min_width=100)
        

        # load the config again to apply window-related options
#        self.config.load()
#        self.config.hidemouse = False
#        self.set_hidemouse(False)


    def set_dictionary(self, dictionary):
        self.worddict = self.loaddictionarygraphically(dictionary)
        self.wordlistprefix = None
        self.internalaction(lambda: self.updatewordbasic(
            self.codetotext(self.getprefixcode())))
        self.updatewordlist()

    def set_hidemouse(self, hide):
        if hide:
            self.buttontable.window.set_cursor(self.nullcursor)
        else:
            self.buttontable.window.set_cursor(self.normalcursor)
                                           

    def loaddictionarygraphically(self, filename):
        messagedialog = gtk.MessageDialog()
        messagedialog.set_title("Tapir 1.0")
        messagedialog.set_markup("Loading dictionary '%s' ..." % filename)
        messagedialog.set_position("center")
        messagedialog.show_all()
        gtkdowork()
        
        worddict = self.loaddictionary(filename)

        messagedialog.destroy()

        return worddict
       
    def symbolcode(self, symbol):
        return self.codeforsymbol.get(symbol, "0")

    def isindividual(self, code):
        return self.individualmode or (code in ["0", "1"])

    def widget(self, name):
        return gtk.glade.XML.get_widget(self.tree, name)

    def save(self):
        self.commitprefix()
        self.config.save()

    def getdict(self):
        if self.individualmode:
            return self.spelldict
        else:
            return self.worddict
        
    def preparedictionary(self):
        self.symbolsforcode = [" .,!?", ".,!?", "abc", "def", "ghi",
                               "jkl", "mno", "pqrs", "tuv", "wxyz"]


        self.spelldict = {"":[""]}
        for code, symbols in listitems(self.symbolsforcode):
            self.spelldict[str(code)] = list(symbols)
        
        self.codeforsymbol = {}
        for code in range(len(self.symbolsforcode)):
            for symbol in self.symbolsforcode[code]:
                self.codeforsymbol[symbol] = str(code)

    def loaddictionary(self, filename):
        worddict = {"0" : self.spelldict["0"],
                    "1" : self.spelldict["1"]}
        
        for line in file(filename):
            words = line.split()
            worddict[words[0][1:]] = words[1:]

        return worddict


    def codetotext(self, code, dict=None):         # weird
        dict = dict or self.getdict()

        if (code in dict) and (len(dict[code]) > 0):
            return dict[code][0]

        if len(code) == 0:
            return ""
        
        for i in reversed(range(1, len(code))):
            if code[:i] in dict:
                return dict[code[:i]][0][:i] + \
                       self.codetotext(code[i-1:], self.suffixdict)[1:]

#                       "?" * (len(code) - i)


        raise "impossible"
        

    def texttocode(self, text):
        return "".join([self.symbolcode(symbol) for symbol in text])

    def getprefixtext(self):
        return marktoiter(self.start).get_visible_text(marktoiter(self.end))

    def getprefixcode(self):
        return self.texttocode(
            marktoiter(self.start).get_visible_text(marktoiter(self.cursor)))

    def updatewordlist(self):
        prefixcode = self.getprefixcode()
        prefixlen = len(prefixcode)
        
        if prefixcode == self.wordlistprefix:
            return

        self.iterhistory = []
        self.liststore.clear()
        for word in self.getdict().get(prefixcode, []):
            if len(word) > prefixlen:
                dashedword = word[:prefixlen] + "-" + word[prefixlen:]
            else:
                dashedword = word
            self.liststore.append([dashedword])
        
        self.selectword(self.getprefixtext())

        self.wordlistprefix = prefixcode
            
    def prefixesmatch(self, code, text):
        if len(code) <= len(text):
            for i in range(len(code)):
                if self.symbolcode(text[i]) != code[i]:
                    return False
            return True
        else:
            return False
        

    def getiters(self):
        return [marktoiter(mark) for mark in self.start, self.cursor, self.end]

    def getoffsets(self):
        return [iter.get_offset() for iter in self.getiters()]

    def removetags(self):
        startiter, cursoriter, enditer = self.getiters()

        self.textbuf.remove_tag(self.badpretag, startiter, enditer)
        self.textbuf.remove_tag(self.pretag, startiter, enditer)
        self.textbuf.remove_tag(self.posttag, startiter, enditer)
        
    def applytags(self):
        startiter, cursoriter, enditer = self.getiters()
        self.textbuf.apply_tag(self.posttag, cursoriter, enditer)
        if self.getcurrentworditer():
            self.textbuf.apply_tag(self.pretag, startiter, cursoriter)
        else:
            self.textbuf.apply_tag(self.badpretag, startiter, cursoriter)
            

        # This is the weirdest thing is this program, it just
        # reapplies the tag applied above, but without this, the
        # program magically stops working.  Actually, the real
        # solution to this problem is to prevent the wordlist from
        # getting focus.
        self.textbuf.apply_tag(self.posttag, cursoriter, enditer)

        self.textview.grab_focus()

    def setmarks(self, startiter, cursoriter, enditer):
        if startiter:
            self.textbuf.move_mark(self.start, startiter)
        if cursoriter:
            self.textbuf.place_cursor(cursoriter)
        if enditer:
            self.textbuf.move_mark(self.end, enditer)

    def say(self, text):
	try:
            self.speech.write('(SayText "%s")' % text)
            self.speech.flush()
	except:
	    pass

    def commitprefix(self):
        startiter, cursoriter, enditer = self.getiters()
        
        if self.speech:
            if enditer.ends_word():
                previter = enditer.copy()
                previter.backward_word_start()
                self.say(previter.get_text(enditer))

            # We do not use enditer.ends_sentence here because it
            # returns True at the end of the text, even if there is no
            # full stop.
            if startiter.get_text(enditer) == ".":
                previter = enditer.copy()
                previter.backward_sentence_start()
                self.say(previter.get_text(enditer))
                
        self.setmarks(enditer, enditer, enditer)

    def settext(self, text):
        startiter, cursoriter, enditer = self.getiters()
        self.textbuf.delete(startiter, enditer)
        self.textbuf.insert(startiter, text)

    def appendcode(self, code):
        prefixcode = self.getprefixcode()

        if self.isindividual(code) or \
           ((len(prefixcode) > 0) and self.isindividual(prefixcode[-1])):
            self.commitprefix()

        prefixcode = self.getprefixcode() + code
#         prefixtext = self.getprefixtext()

#         if self.prefixesmatch(prefixcode, prefixtext):
#             cursoriter = marktoiter(self.cursor)
#             cursoriter.forward_char()
#             self.setmarks(None, cursoriter, None)
#         else:
        self.updatewordbasic(self.codetotext(prefixcode), 1)
            
    def internalaction(self, function, block=False):
        if not self.internal:
            try:
                self.internal = True
                self.removetags()
                function()
            finally:
                self.updatewordlist()
                self.applytags()
                self.textview.scroll_mark_onscreen(self.cursor)
                self.internal = False
        elif not block:
            function()

    def enable(self):
        cursoriter = marktoiter(self.cursor)
        if not cursoriter.is_start():
            previter = cursoriter.copy()
            previter.backward_char()

            if not self.isindividual(self.symbolcode(previter.get_char())):
                previter = cursoriter.copy()
                previter.backward_word_start()

            self.setmarks(previter, None, cursoriter)    

    def removechar(self):
        startiter, cursoriter, enditer = self.getiters()
        if not startiter.equal(cursoriter):
            prefixcode = self.getprefixcode()
            self.updatewordbasic(self.codetotext(prefixcode[:-1]), -1)
            
        startiter, cursoriter, enditer = self.getiters()
        if startiter.equal(cursoriter):
            self.textbuf.delete(startiter, enditer)
            self.enable()
            
    def disable(self, cursoriter=None):
        if not cursoriter:
            cursoriter = marktoiter(self.cursor)
        self.setmarks(cursoriter, None, cursoriter)
            
    def onmarkset(self, textbuf, iter, mark, data=None):
        if mark.get_name() == "insert":
            self.internalaction(lambda : self.disable(iter), True)

    def onchanged(self, textbuf):
        self.internalaction(lambda : self.disable(), True)

    def quit(self):
        self.layout.setlayout("confirmquit")
            
    def onquit(self):
        self.copytoclipboard(self)
        self.save()
	gtk.main_quit()
        
    def selectword(self, word):
        model = self.wordlist.get_model()
        iter = model.get_iter_root()
        while iter is not None:
            if self.getwordatiter(iter) == word:
                self.wordlist.set_cursor(model.get_path(iter))
            iter = model.iter_next(iter)

    def getcurrentworditer(self):
        model = self.wordlist.get_model()
        path, column = self.wordlist.get_cursor()
        if path is not None:
            return model.get_iter(path)
        else:
            return None

    def getcurrentword(self):
        iter = self.getcurrentworditer()
        if iter:
            return self.liststore.get_value(iter, 0).replace("-","")
        else:
            return None

    def updatewordbasic(self, text, cursormove=0):
        startoff, cursoroff, endoff = self.getoffsets()
        self.settext(text)
        
        startiter = self.textbuf.get_iter_at_offset(startoff)
        cursoriter = self.textbuf.get_iter_at_offset(cursoroff + cursormove)
        self.setmarks(startiter, cursoriter, None)
        

    def updateword(self, text=None):
        if text is None:
            text = self.getcurrentword()

        self.updatewordbasic(text)

        self.say(text)

    def set_fontname(self, fontname):
        fontdesc = pango.FontDescription(fontname)
        if fontdesc:
            recursewidgets(self.mainwindow,
                           lambda widget: widget.modify_font(fontdesc))

    def set_spacing(self, spacing):
        self.spacingscale.set_value(spacing)
        self.applyspacing()

    def get_spacing(self):
        return int(self.spacingscale.get_value())

    def applyspacing(self):
        spacing = self.get_spacing()
        self.buttontable.set_row_spacings(spacing)
        self.buttontable.set_col_spacings(spacing)        

    # reactions to user actions

    def on_BackspaceButton_pressed(self, button):
        if len(self.iterhistory) > 0:
            self.internalaction(self.prev)
        else:
            self.internalaction(self.removechar)

    def on_MainWindow_delete_event(self, widget, event):
        self.onquit()

#     def on_SaveQuitButton_pressed(self, button):
#         self.onquit()

    def on_NextButton_pressed(self, button):
        if self.individualmode:
            self.internalaction(lambda : self.commitprefix())
        else:
            self.internalaction(self.next)

    def prev(self):
        iter = self.iterhistory.pop()
        self.wordlist.set_cursor(self.liststore.get_path(iter))

    def next(self):
        iter = self.getcurrentworditer() 
        if iter:
            self.iterhistory.append(iter)
	next = iter and self.liststore.iter_next(iter)
        if next is None:
            next = self.liststore.get_iter_root()
        if next is not None:
            self.wordlist.set_cursor(self.liststore.get_path(next))
        

    def getwordatiter(self, iter):
        return self.liststore.get_value(iter, 0).replace("-","")

    def populatebuttons(self):
	iter = self.getcurrentworditer() or self.liststore.get_iter_root()

        if not iter:
            self.layout.setlayout("main")
            return

        for widget in self.layout.widgets:
            if iter and (getlabel(widget) == ""):
                setlabel(widget, self.getwordatiter(iter))
                iter = self.liststore.iter_next(iter)

        iter = iter or self.liststore.get_iter_root()

        if iter:
            self.internalaction(lambda:
                self.wordlist.set_cursor(self.liststore.get_path(iter)))

    def acceptword(self, word):
        if word:
            self.internalaction(lambda : self.updateword(word))
            self.internalaction(lambda : self.appendcode("0"))

    def accepttitle(self, button):
        self.acceptword(getlabel(button))

    def getindividualcode(self):
        if self.individualmode:
            prefix = self.getprefixcode()
            if len(prefix) == 1:
                return prefix

    def onbuttoncode(self, numcode):
        code = str(numcode)
        if self.getindividualcode() == code:
            self.internalaction(self.next)
        else:
            self.internalaction(lambda : self.appendcode(code))

    def on_Button_pressed(self, button):
        self.layout.action(button)
#         self.onbuttoncode(self.getbuttoncode(button))
        
    def on_ToggleButton_pressed(self, button):
	pass

    def on_WordList_cursor_changed(self, treeview):
        if self.internal:
            self.internalaction(self.updateword)
        else:
            self.acceptword(self.getcurrentword())

    def on_FontButton_pressed(self, button):
        fontdialog = gtk.FontSelectionDialog("Select Font")

        if self.config.fontname:
            fontdialog.set_font_name(self.config.fontname)

        if fontdialog.run() == gtk.RESPONSE_OK:
            self.config.fontname = fontdialog.get_font_name()
            
        fontdialog.destroy()

    def on_SpacingScale_value_changed(self, scale):
        self.applyspacing()


    def movecursor(self, function):
        iter = marktoiter(self.textbuf.get_insert())
        function(iter)
        self.textbuf.place_cursor(iter)


    def copytoclipboard(self, delete=False):
        start, end = self.textbuf.get_start_iter(), self.textbuf.get_end_iter()
        text = self.textbuf.get_text(start, end)
        if len(text.strip()) > 0:
            gtk.clipboard_get("CLIPBOARD").set_text(text)
            gtk.clipboard_get("PRIMARY").set_text(text) # ???
            if delete:
                self.textbuf.delete(start, end)

    def pastefromclipboard(self):
        text = gtk.clipboard_get("CLIPBOARD").wait_for_text()
        if text:
            self.textbuf.insert_at_cursor(text)

    def reenable(self):
        self.disable()
        self.enable()        

    def switchmode(self):
        self.disable()
        self.individualmode = not self.individualmode
        self.enable()        

    def set_windowsize(self, size):
        x, y = size
        self.mainwindow.resize(x, y)
#        self.mainwindow.set_size_request(x, y)

    def get_windowsize(self):
        return self.mainwindow.get_size()
        

myapp = MyApp()
gtk.main()
